<?php

use Intervention\Image;

/**
 * GenticsImageStore (PHP version)
 * Module that allows you to crop & resize images on the fly, during page load
 * and caches the resized images on the filesystem for later use.
 * The images can be loaded from the content repository (DynamicContentSource),
 * filesystem or any PHP fopen wrappers (HTTP, FTP, ...).
 *
 * Requires PHP 5.3 or higher
**/
class GenticsImageStoreModule extends CWebModule
{
    /**
     * @var string  Default image driver, should be either "imagick" or "gd"
     */
    public $imageDriver = "imagick";

    /**
     * @var array  List of supported resize modes
     */
    public $resizeModes = array(
        'prop',
        'force',
        'smart'
    );

    /**
     * @var array  Cache Properties
     */
    public $cache = array(
        'folder'         => null,
        'maxAge'         => 0,  // 7 days
        'maxSize'        => 10737418240, // 10 GB
        'instantCleanup' => true,
    );

    /**
     * @var bool true when the images should cropped and resized on the first request to the image
     */
    public $onDemandResize = false;

    /**
     * @var string the file ending of the images containing the information for the on demand resized images
     */
    public $onDemandFileEnding = '.info';

    /**
     * @var Intervention\Image\ImageManager  An ImageManager instance
     */
    protected $imageManager = null;

    /**
     * @var boolean Set to true when the cache cleanup has run once
     */
    protected $cacheCleanedupOnce = false;


    /**
     * Initialize the module
     * @throws Exception when the cache directory could not be created
     */
    public function init()
    {
        // Note: The Intervention namespace is loaded in main.php

        // Don't use Imagick if the class is unknown
        if ($this->imageDriver === "imagick" && !class_exists("Imagick", false)) {
            $this->imageDriver = "gd";
        }

        // Create an image manager instance with Image Magick (preferred) or GD
        if ($this->imageDriver === "imagick") {
            $this->imageManager = new Image\ImageManager(array('driver' => 'imagick'));
        } else {
            $this->imageManager = new Image\ImageManager(array('driver' => 'gd'));
        }

        if (empty($this->cache['folder'])) {
            $contentSource = $this->getContentSourceModule()->getContentSource();

            if ($contentSource instanceof DynamicContentSource) {
                $this->cache['folder'] .= $contentSource->cacheFolder;
            } else {
                $this->cache['folder'] .= $contentSource->contentFolder;
            }

            $this->cache['folder'] .= "/GenticsImageStore";
        }

        if (!$this->ensureFolderExists($this->cache['folder'])) {
            throw new Exception("Could not create directory \"" . $this->cache['folder'] . "\"");
        }
    }

    /**
     * Get the initialized instance of the image manager.
     * Can be used to resize images.
     * @return Intervention\Image\ImageManager
     */
    public function getImageManager()
    {
        return $this->imageManager;
    }

    /**
     * Path to the cache folders where transformed images are cached
     * @return string  The path
     */
    public function getCacheFolder()
    {
        return $this->cache['folder'];
    }

    /**
     * Get the cache options
     * @return array Cache options
     */
    public function getCacheOptions()
    {
        return $this->cache;
    }

    /**
     * Given a filename of a cached file, it returns the full path to
     * this image in the cache directory. If on demand resize is active 
     * and resize info file is given, the image gets resized.
     * @param string $fileName        Filename of the image
     * @param bool $checkExistence  Whether to check if the file actually exists
     * @return string  The image path or null
     */
    public function getImagePath($fileName, $checkExistence=true)
    {
        $cacheFolder = $this->getCacheFolder();
        $imageFilePath = $cacheFolder . "/" . $fileName;

        if ($checkExistence && !file_exists($imageFilePath)) {
            // if on demand resize is active we have to check if we can find 
            // a resize info file for the requested filename
            if ($this->onDemandResize) {
                $infoFilePath = $imageFilePath . $this->onDemandFileEnding;
                if (file_exists($infoFilePath)) {
                    $data = unserialize(file_get_contents($infoFilePath));
                    // ensure the integrity of the unserialized data before using it to fetch and resize the image
                    if (isset($data['id']) && isset($data['resizeOptions']) && isset($data['cropOptions']) && isset($data['publishPath'])) {
                            $this->doRepositoryResize($data['id'], $data['resizeOptions'], $data['cropOptions'], $data['publishPath']);
                    }
                    // if we reach this and all went well the file should available in $imageFilePath
                    // to prevent multiple failing requests on the same image url we remove the infoFile
                    unlink($infoFilePath);
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }

        return $imageFilePath;
    }

    /**
     * Asks the content repository for the relative path to the given object.
     * Only works if DCR is enabled.
     * @param integer $objectId ID of the object in the CR, without the type
     * @param string|null $publishPath optional the publish path of the image
     * @return string Relative URL/path to the file (publish path)
     */
    protected function getRepositoryObjectFilePath($objectId, $publishPath = null)
    {
        $content = $this->getContentSourceModule()->getContentById(
                "10008." . $objectId, $publishPath);
        if ($content === null) {
            return null;
        }

        return $content->getFile();
    }

    /**
     * Given a content repository publish path, it crops and/or resizes the
     * image and caches the result if not cached already.
     * @param string $url            Relative publish path
     * @param array $resizeOptions
     * @param array $cropOptions
     * @return string Relative URL/path to the file (publish path)
     */
    public function transformImageByFileUrl($url, $resizeOptions = array(),
            $cropOptions = array())
    {
        if (empty($url)) {
            return null;
        }

        // file_exists doesn't work with URLs, so it's one of the rare
        // legitimate use cases of the @ operator here to surpress the warning.
        $content = @file_get_contents($url);
        if ($content === false) {
            return null;
        }

        $urlComponents = parse_url($url);
        $filePath = trim($urlComponents['path']);

        $outputFileName = $this->getUniqueFilename(basename($filePath),
                array($url, $resizeOptions, $cropOptions, md5($content)));

        return $this->transformImage($content, $outputFileName,
                $resizeOptions, $cropOptions);
    }

    /**
     * Given a content repository object ID, it crops and/or resizes the
     * image and caches the result if not cached already.
     * @param string $id             ID of the content repository object
     *                               
     * @param array  $resizeOptions  Resize options
     * @param array  $cropOptions    Crop options
     * @param string $publishPath    Optional: Path to the file in the node,
     *                               this is used for optimization only to 
     *                               prevent unnecessary CR lookups, can be null
     * @return string Relative URL/path to the file (publish path)
     */
    public function transformImageByRepositoryId($id,
            $resizeOptions, $cropOptions, $publishPath = null)
    {
        // check if on demand resize is active and we can resize on demand (only when a publish path is available)
        if ($this->onDemandResize && !empty($publishPath)) {
            return $this->doOnDemandResize($id, $resizeOptions, $cropOptions, $publishPath);
        } else {
            return $this->doRepositoryResize($id, $resizeOptions, $cropOptions, $publishPath);
        }
    }

    /**
     * Given a content repository object ID, it crops and/or resizes the
     * image and caches the result if not cached already.
     * @param string $id             ID of the content repository object
     *
     * @param array  $resizeOptions  Resize options
     * @param array  $cropOptions    Crop options
     * @param string $publishPath    Optional: Path to the file in the node,
     *                               this is used for optimization only to
     *                               prevent unnecessary CR lookups, can be null
     * @return string Relative URL/path to the file (publish path)
     */
    public function doRepositoryResize($id, $resizeOptions, $cropOptions, $publishPath = null)
    {

        $originalImagePath = $this->getRepositoryObjectFilePath($id, $publishPath);
        if ($originalImagePath === null) {
            return null;
        }

        // use the publish path to generate the unique filenames when doing on demand resize
        if (!empty($publishPath) && $this->onDemandResize) {
            $uniquePath = $publishPath;
        } else {
            $uniquePath = $originalImagePath;
        }
        $outputFileName = $this->getUniqueFilename(basename($uniquePath)
            , array($uniquePath, $resizeOptions, $cropOptions));

        // Delete the cache file, if the original image on the filesystem is newer
        if (file_exists($outputFileName) &&
            filemtime($originalImagePath) > filemtime($outputFileName)) {
            unlink($outputFileName);
        }

        return $this->transformImage($originalImagePath,
            $outputFileName, $resizeOptions, $cropOptions);
    }

    /**
     * Crops and resizes a image
     * @param mixed $imageData       Can be a binary image content, a file path
     *                                or other things, for more imformation see
     *                                the documentation of Intervention\Image\ImageManager::make
     * @param string $outputFileName  The full path to the resulting output file
     * @param array $resizeOptions    Resize options
     * @param array $cropOptions      Crop options
     * @return string the URL that can be rendered in a page
     * @throws Exception When the cache directory can't be created
     */
    public function transformImage($imageData, $outputFileName,
            $resizeOptions = array(), $cropOptions = array())
    {
        $imageManager = $this->getImageManager();
        $cacheFolder = $this->getCacheFolder();
        $outputFilePath = $cacheFolder . "/" . $outputFileName;

        if (!file_exists($outputFilePath)) {
            // Cleanup the whole cache, if enabled and hasn't run previously
            $cacheOptions = $this->getCacheOptions();
            if (!$this->cacheCleanedupOnce
                    && isset($cacheOptions['instantCleanup'])
                    && $cacheOptions['instantCleanup']) {
                $this->cleanupCache();
            }

            // Resize image
            $image = $imageManager->make($imageData);

            if (is_array($cropOptions) && !empty($cropOptions)) {
                $this->cropImage($image, $cropOptions);
            }

            if (is_array($resizeOptions) && !empty($resizeOptions)) {
                $this->resizeImage($image, $resizeOptions);
            }

            $image->save($outputFilePath);
        }

        // The urlManager will automatically create a pretty URL for us
        // without "show" and without query parameter.
        return Yii::app()->getUrlManager()->createUrl(
            "genticsimagestore/image/show", array('fileName' => $outputFileName));
    }


    /**
     * On demand crop and resize of a image. The image will be fetched and resized
     * when its requested for the first time. In the meantime a (text-)file containing
     * all the necessary information for fetching and resizing the image will be stored
     * in the image cache.
     *
     * @param string $id             ID of the content repository object
     *
     * @param array  $resizeOptions  Resize options
     * @param array  $cropOptions    Crop options
     * @param string $publishPath    Path to the file in the node,
     *                               this is used for optimization only to
     *                               prevent unnecessary CR lookups, can be null
     * @return string the URL that can be rendered in a page
     * @throws Exception When the cache directory can't be created
     */
    public function doOnDemandResize($id, $resizeOptions, $cropOptions, $publishPath)
    {
        $outputFileName = $this->getUniqueFilename(basename($publishPath)
            , array($publishPath, $resizeOptions, $cropOptions));

        $cacheFolder = $this->getCacheFolder();
        $outputFilePath = $cacheFolder . "/" . $outputFileName;
        $infoFilePath = $outputFilePath . $this->onDemandFileEnding;
        // check if the image was already resized or if an on demand info file was already generated for this file
        if (!file_exists($outputFilePath) && !file_exists($infoFilePath)) {
            // Cleanup the whole cache, if enabled and hasn't run previously
            $cacheOptions = $this->getCacheOptions();
            if (!$this->cacheCleanedupOnce
                && isset($cacheOptions['instantCleanup'])
                && $cacheOptions['instantCleanup']) {
                $this->cleanupCache();
            }
            $data = array(
                'id' => $id,
                'publishPath' => $publishPath,
                'resizeOptions' => $resizeOptions,
                'cropOptions' => $cropOptions
            );
            file_put_contents($infoFilePath, serialize($data));
        }

        // The urlManager will automatically create a pretty URL for us
        // without "show" and without query parameter.
        return Yii::app()->getUrlManager()->createUrl(
            "genticsimagestore/image/show", array('fileName' => $outputFileName));
    }

    /**
     * Based on a filename and a data array, it creates a unqiue filename with
     * a MD5 hash
     * @param string $fileName  File name
     * @param array $data       You can put any data into this array, it will
     *                         be used to generate the MD5 checksum
     * @return string  Unique file name
     */
    public function getUniqueFilename($fileName, array $data)
    {
        return md5(serialize($data)) . "-" . $fileName;
    }

    /**
     * prop used for proportional resizing.
     *      The aspect ratio of the image will be preserved. When target width
     *      and height do not match the original image proportions, the lower
     *      value (height, width) will be used for resizing.
     * force used for forced resizing.
     *      The aspect ratio of the original image will not be preserved when
     *      this mode is used.
     * smart used for automatic resizing and cropping of images.
     * @param Intervention\Image\Image $image    The image to manipulate
     * @param array                           $options  Resize options
     */
    protected function resizeImage($image, array $options)
    {
        $newWidth  = null;
        $newHeight = null;
        $mode      = "smart";

        // Takeover the mode, if exists
        if (!empty($options['mode'])
                && in_array($options['mode'], $this->resizeModes)) {
            $mode = $options['mode'];
        }

        // Make sure the width and height is always either null (auto)
        // or an integer value greater 0
        if ($options['width'] !== "auto" && $options['width'] > 0) {
            $newWidth = $options['width'];
        }

        if ($options['height'] !== "auto" && $options['height'] > 0) {
            $newHeight = $options['height'];
        }

        if ($mode === "smart") {
            // In smart mode, the image should be cropped and resized
            // automatically and the image should be centered
            if ($newWidth == null) {
                // If auto or 0 was passed, we use the original image width
                $image->fit($image->width(), $newHeight);
            } else if ($newHeight == null) {
                // If auto or 0 was passed, we use the original image height
                $image->fit($newWidth, $image->height());
            } else {
                $image->fit($newWidth, $newHeight);
            }
        } else {
            $constraintFunction = function ($constraint) use ($mode) {
                if ($mode === "prop") {
                    // Keep the ratio of the original image
                    $constraint->aspectRatio();
                }
            };

            $image->resize($newWidth, $newHeight, $constraintFunction);
        }
    }

    /**
     * Crops a Intervention image
     * @param Intervention\Image\Image $image    The image to manipulate
     * @param array                    $options  Crop options
     */
    protected function cropImage($image, $options)
    {
        $newWidth  = null;
        $newHeight = null;
        $positionX = null;
        $positionY = null;
        $mode      = "smart";

        if (!empty($options['mode'])
                && in_array($options['mode'], $this->resizeModes)) {
            $mode = $options['mode'];
        }

        if (!empty($options['width'])
                && $options['width'] !== "auto" && $options['width'] > 0) {
            $newWidth = $options['width'];
        }

        if (!empty($options['height'])
                && $options['height'] !== "auto" && $options['height'] > 0) {
            $newHeight = $options['height'];
        }

        if (!empty($options['x'])
                && $options['x'] !== "auto" && $options['x'] > 0) {
            $positionX = $options['x'];
        }

        if (!empty($options['y'])
                && $options['y'] !== "auto" && $options['y'] > 0) {
            $positionY = $options['y'];
        }

        if ($newWidth === null && $newHeight === null
                && $positionX === null && $positionY === null) {
            return;
        }

        $image->crop($newWidth, $newHeight, $positionX, $positionY);
    }

    /**
     * Creates the given directory path recursively
     * @param string $path  Folder path to create
     * @return boolean True if it already existed or if it was created successfully
     */
    protected function ensureFolderExists($path)
    {
        if (!file_exists($path)) {
            return mkdir($path, 0770, true);
        }

        return true;
    }

    /**
     * Delete all cached images
     * @return void
     */
    public function clearCache()
    {
        $cacheFolder = $this->getCacheFolder();

        if (empty($cacheFolder) || !file_exists($cacheFolder)) {
            return;
        }

        $files = glob($cacheFolder . "/*");
        foreach($files as $file) {
            if (is_file($file)) {
                unlink($file);
            }
        }
    }

    /**
     * Get the content source module
     * @return ContentSourceModule the content source module
     */
    public function getContentSourceModule() {
        return Yii::app()->getModule('contentSource');
    }

    /**
     * Delete too old cache files and the oldest files if the cache is too big,
     * according to the user settings.
     * @return void
     */
    public function cleanupCache()
    {
        $this->cacheCleanedupOnce = true;

        /*
         * 'maxAge'         => 604800,
         * 'maxSize'        => 10737418240,
         * 'instantCleanup' => true,
         */
        $cacheFolder = $this->getCacheFolder();
        $cacheOptions = $this->getCacheOptions();

        if (empty($cacheFolder) || !file_exists($cacheFolder)) {
            return;
        }

        $maxAge  = null;
        $maxSize = null;

        if (!empty($cacheOptions['maxAge'])) {
            $maxAge = $cacheOptions['maxAge'];
        }

        if (!empty($cacheOptions['maxSize'])) {
            $maxSize = $cacheOptions['maxSize'];
        }

        if (empty($maxAge) && empty($maxSize)) {
            return;
        }

        $cacheFiles = glob($cacheFolder . "/*");

        if (empty($cacheFiles)) {
            return;
        }

        // Fetch all file modification times and calculate filesize sum
        $fileModificationTimes = array();
        $fileSizes             = array();
        $sumFileSizes = 0;
        foreach ($cacheFiles as $cacheFile) {
            $fileModificationTimes[$cacheFile] = filemtime($cacheFile);

            if ($maxSize > 0) {
                $fileSize = filesize($cacheFile);
                $sumFileSizes += $fileSize;
                $fileSizes[$cacheFile] = $fileSize;
            }
        }

        // Sort the files by modification time
        usort($cacheFiles, function ($fileA, $fileB) use($fileModificationTimes)
        {
            return $fileModificationTimes[$fileA] - $fileModificationTimes[$fileB];
        });

        // Delete files that are too old (maxAge setting)
        $oldestTime = time() - $maxAge;
        foreach ($cacheFiles as $cacheFile) {
            if ($maxAge > 0 && $fileModificationTimes[$cacheFile] < $oldestTime) {
                unlink($cacheFile);

                if ($maxSize > 0) {
                    $sumFileSizes -= $fileSizes[$cacheFile];
                }
            }
        }

        // If the cache is still too big, delete the oldest files until
        // the cache size is acceptable (maxSize setting)
        if ($maxSize > 0 && $sumFileSizes > $maxSize) {
            foreach ($cacheFiles as $cacheFile) {
                unlink($cacheFile);
                $sumFileSizes -= $fileSizes[$cacheFile];

                if ($sumFileSizes <= $maxSize) {
                    break;
                }
            }
        }
    }
}
